# =============================================================================
# exceptions
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


class GeneralError(Exception):
    """General error"""
    def __init__(self, message=None, host=None):
        """The class constructor.

        Args:
            message (str): Custom message to be passed to the exceptions. Defaults to *None*.
                If *None* then the general class *__doc__* is used.
            host (str): Custom string which can be used to enhance the exception message by adding the "`host`: "
                prefix to the message string. Defaults to *None*. If `host` is *None* then message stays unchanged.
        """
        self.message = message
        self.hostname = str(host)

    def __str__(self):
        message = self.message or self.__class__.__doc__
        return "{}: {}".format(self.hostname, message) if self.hostname else message


class InvalidHopInfoError(GeneralError):
    """Invalid device connection parameters"""
    pass


class ConnectionError(GeneralError):
    """General connection error"""
    pass


class ConnectionAuthenticationError(ConnectionError):
    """Connection authentication error"""
    pass


class ConnectionTimeoutError(ConnectionError):
    """Connection timeout error"""
    pass


class CommandError(GeneralError):
    """Command execution error"""
    def __init__(self, message=None, host=None, command=None):
        """The class constructor.

        Args:
            message (str): Custom message to be passed to the exceptions. Defaults to *None*.
                If *None* then the general class *__doc__* is used.
            host (str): Custom string which can be used to enhance the exception message by adding the "`host`: "
                prefix to the message string. Defaults to *None*. If `host` is *None* then message stays unchanged.
            command (str): Custom string which can be used enhance the exception message by adding the
                "`command`" suffix to the message string. Defaults to *None*. If `command` is *None* then the message
                stays unchanged.
        """
        GeneralError.__init__(self, message, host)
        self.command = command

    def __str__(self):
        message = self.message or self.__class__.__doc__
        message = "{}: '{}'".format(message, self.command) \
            if self.command else message
        message = "{}: {}".format(self.hostname, message) \
            if self.hostname else message
        return message


class CommandSyntaxError(CommandError):
    """Command syntax error"""
    pass


class CommandTimeoutError(CommandError):
    """Command timeout error"""
    pass



# =============================================================================
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import sys
import os
import re
import logging
import time

from hopinfo import make_hop_info_from_url
from controllers.pexpect_ctrl import Controller
from condoor.utils import delegate

from pexpect import TIMEOUT
from condoor.controllers.fsm import FSM, action

from condoor.exceptions import CommandTimeoutError, ConnectionError, ConnectionTimeoutError, CommandError, \
    CommandSyntaxError, ConnectionAuthenticationError, GeneralError, InvalidHopInfoError

__version__ = '0.1.3'

"""
This is a python module providing access to Cisco devices over Telnet and SSH.

"""

__all__ = ['make_connection_from_urls', 'Connection', 'FSM', 'TIMEOUT', 'action',
           'CommandTimeoutError', 'ConnectionError', 'ConnectionTimeoutError', 'CommandError',
           'CommandSyntaxError', 'ConnectionAuthenticationError']

drivers = {
    "ASR9K": ["ASR9K", "CRS", "NCS6K", "NCS4K", "CRS", "NCS5K", "NCS5500", "NCS1K"],
    "IOS": ["ASR900"],
    "NX-OS": ["N9K"],
    "generic": ["generic"]

}

os_names = {
    'IOS': 'IOS',
    'XR': 'IOS XR',
    'eXR': 'IOS XR 64 bit',
    'XE': 'IOS XE',
    'NX-OS': 'NX-OS',

}


@delegate("_driver", ("reload", "send", "enable", "run_fsm"))
class Connection(object):
    """This is the main class interface for Condoor. Use this class to create
    a connection session, discover and control the remote device."""

    def __init__(self, name, urls, log_dir=None, log_level=logging.DEBUG, log_session=True, account_manager=None):
        """This is the constructor. The *hostname* parameter is a string representing the name of the device.
        It is used mainly for verbose logging purposes.

        The *urls* parameter represents the list of urls used to define the connection way to the
        chain of devices before reaching the target device. It allows to have single device connection
        over multiple jumphosts. For example::

            ["ssh://<user>:<password>@<host>", "telnet://<user>:<password>@<host>:<port>"]

        Currently there are two protocols supported: SSH and TELNET. The *port* is optional and if not specified
        the well known default protocol port is used.

        An IOS/IOS XE devices may require a privileged level password (enable password). The url parser treats
        the rest of the url (the path part + query + fragment part) as an privileged password. For example::

            ["ssh://<user>:<password>@<host>", "telnet://<user>:<password>@<host>/<enable_password>"]

        The *host* could be either the hostname or ip address.

        There could be more devices in the list representing the single or multiple jumphosts on the path
        to the target device. The condoor connects to them in a sequence starting the next connection from the
        previous host unless reaching the target device. The last url in the list is always the target device.

        The *urls* parameter can be passed as a list of the list of urls. In that case condoor treats it as
        a multiple alternative connections to the same device. During the connection or discovery phase condoor
        starts with the first connection from the list and then move sequentially to next connection list in case of
        the previous connection failure or timeout.
        This process stops when the current connection to the target devices is successful.
        The first successful connection is latched and stored in the connection class and next time when connection
        to device is being reestablished the last successful connection list is used.
        This feature is useful for providing the console connection to two processor cards of the same device.
        It could be also used if the device has two alternative Mgmt interfaces with different IP addresses.
        For example::

                [["ssh://<user>:<password>@<host>", "telnet://<user>:<password>@<ts1:2015>"],
                 ["ssh://<user>:<password>@<host>", "telnet://<user>:<password>@<ts1:2016>"]]

        In above example the *urls* parameter represents two alternative connection lists. First goes through
        terminal server on port 2015 and second goes through the same terminal server on port 2016.
        Both connections uses the same jumphost.

        The *log_dir* parameter contains a string representing the full path to the logging directory.
        The logging directory can store the device session log and the detailed Condoor debug log.
        If there is no *log_dir* specified the current directory is used. The possible logging levels are as follows::

            NOTSET=0
            DEBUG=10
            INFO=20
            WARN=30
            ERROR=40
            CRITICAL=50

        The default is DEBUG. If the *log_level* is set to 0 then no logging file is created.  The *log_session*
        parameters defines whether the device session log is created or not.

        """

        self._driver = None
        self._name = name
        self._hostname = name

        # normalise the urls to the list of the lists of str (for multiple console connections to the device)
        if isinstance(urls, list):
            if urls:
                if isinstance(urls[0], list):
                    # multiple connections (list of the lists)
                    self._urls = urls
                elif isinstance(urls[0], str):
                    # single connections (make it list of the lists)
                    self._urls = [urls]
            else:
                raise GeneralError("No target host url provided.")
        elif isinstance(urls, str):
            self._urls = [[urls]]

        nodes = []
        for index, target in enumerate(self._urls):
            nodes.insert(index, list())
            for url in target:
                nodes[index].append(make_hop_info_from_url(url))
        self._nodes = nodes
        self._last_driver_index = 0

        self._account_manager = account_manager
        self._platform = 'generic'
        self._os_type = None
        self._os_version = None
        self._family = None
        self._prompt = None
        self._log_dir = log_dir
        self._log_session = log_session
        self.logger = logging.getLogger('condoor')
        self._info = {}
        self._is_console = False

        self._udi = {
            "name": "",
            "description": "",
            "pid": "",
            "vid": "",
            "sn": ""
        }

        if log_level > 0:
            formatter = logging.Formatter('%(asctime)-15s %(levelname)8s: %(message)s')
            if log_dir:
                # Create the log directory.
                if not os.path.exists(log_dir):
                    try:
                        os.makedirs(log_dir)
                    except IOError:
                        log_dir = "./"
                log_filename = os.path.join(log_dir, 'condoor.log')
                handler = logging.FileHandler(log_filename)

            else:
                handler = logging.StreamHandler()
                log_dir = "./"

            handler.setFormatter(formatter)
        else:
            handler = logging.NullHandler()

        self._handler = handler

        self.logger.addHandler(self._handler)
        self.logger.setLevel(log_level)

        try:
            self._session_fd = open(os.path.join(log_dir, 'session.log'), mode="w")
        except IOError:
            self._session_fd = None

    def __del__(self):
        self.logger.removeHandler(self._handler)

    def _set_default_log_fd(self, logfile=None):
        if self._log_session:
            if logfile is None:
                try:
                    self._session_fd = open(os.path.join(self._log_dir, 'session.log'), mode="a+")
                except (IOError, AttributeError):
                    self._session_fd = None
            else:
                self._session_fd = logfile if isinstance(logfile, file) else None

    def _get_driver_name(self):
        for driver_name, families in drivers.iteritems():
            if self._family in families:
                return driver_name
        else:
            return 'generic'

    def _init_driver(self, driver_name='generic'):

        if driver_name == 'generic':
            if self._os_type in ["eXR", "XR"]:
                driver_name = 'ASR9K'  # TODO: change the driver name to IOSXR
            elif self._os_type in ["IOS", "XE"]:
                driver_name = 'IOS'
            elif self._os_type in ['NX-OS']:
                driver_name = 'NX-OS'

        module_str = 'condoor.platforms.%s' % driver_name
        try:
            __import__(module_str)
            module = sys.modules[module_str]
            driver_class = getattr(module, 'Connection')
        except ImportError:
            raise GeneralError("Platform {} not supported".format(self.platform))

        driver = driver_class(
            self._hostname,
            self._nodes[self._last_driver_index],
            Controller,
            self.logger,
            account_manager=self._account_manager
        )
        self._driver = driver

    def _shift_driver(self):
        no_hosts = len(self._nodes)
        del self._driver
        self._last_driver_index += 1
        if self._last_driver_index >= no_hosts:
            self._last_driver_index = 0

        self._init_driver()

    def _detect_console(self):
        try:
            output = self._driver.send("show users")
        except CommandError:
            self.logger.debug("Command 'show users' not suported")
            return False

        for line in output.split('\n'):
            if '*' in line:
                break
        else:
            self.logger.debug("Connection port unknown")
            return False

        if 'vty' in line:
            self.logger.debug("Detected connection to vty")
            return False
        elif 'con' in line or 'tty' in line:  # tty for NX-OS
            self.logger.debug("Detected connection to console")
            return True

        self.logger.debug("Connection port unknown")
        return False

    def _update_device_info(self):
        try:
            show_version = self._driver.send("show version brief", timeout=120)
        except CommandError:
            # IOS Hack - need to check if show version brief is supported on IOS/IOSXE
            show_version = self._driver.send("show version", timeout=120)

        match = re.search("Version (.*?)(?:\[| |$)", show_version, re.MULTILINE)
        if match:
            self._os_version = match.group(1)

        match = re.search("System version: (.*)", show_version, re.MULTILINE)
        if match:
            self._os_version = match.group(1)  # override for NX-OS

        match = re.search("(XR|XE|NX-OS)", show_version)
        if match:
            self._os_type = match.group(1)
        else:
            self._os_type = 'IOS'

        if self._os_type == "XR":
            match = re.search("Build Information", show_version)
            if match:
                self._os_type = "eXR"

        #match = re.search("^[ ?]cisco (.*?) ", show_version, re.MULTILINE)  # NX-OS
        match = re.search("^(  )?cisco (.*?) ", show_version, re.MULTILINE)  # NX-OS
        if match:
            self.logger.debug("Platform string: {}".format(match.group()))
            self._platform = match.group(2)
            _family = match.group(2)
        else:
            self._family = 'generic'
            return

        if _family.startswith("ASR9K"):
            _family = "ASR9K"
        elif _family.startswith("NCS-6"):
            _family = "NCS6K"
        elif _family.startswith("NCS-4"):
            _family = "NCS4K"
        elif _family.startswith("NCS-50"):
            _family = "NCS5K"
        elif _family.startswith("NCS-55"):
            _family = "NCS5500"
        elif _family.startswith("CRS"):
            _family = "CRS"
        elif _family.startswith("ASR-9") and self._os_type == "XE":
            _family = "ASR900"
        elif _family.startswith("Nexus9000") and self._os_type == "NX-OS":
            _family = "N9K"
        elif _family.startswith("NCS1"):
            _family = "NCS1K"

        self._family = _family

    def _update_udi(self):

        if self._os_type in ['XR', 'eXR']:
            cmd = 'admin show inventory chassis'
        elif self._os_type in ['IOS', 'XE', 'NX-OS']:
            cmd = 'show inventory'
        else:
            return self._udi  # do not detect

        # if command not supported return empty uid dict so far
        try:
            show_inventory_chassis = self._driver.send(cmd)
        except CommandError:
            return self._udi

        match = re.search(r"(?i)NAME: (?P<name>.*?),? (?i)DESCR", show_inventory_chassis, re.MULTILINE)
        if match:
            self._udi['name'] = match.group('name').strip('" ,')

        match = re.search(r"(?i)DESCR: (?P<description>.*)", show_inventory_chassis, re.MULTILINE)
        if match:
            self._udi['description'] = match.group('description').strip('" ')

        match = re.search(r"(?i)PID: (?P<pid>.*?),? ", show_inventory_chassis, re.MULTILINE)
        if match:
            self._udi['pid'] = match.group('pid')

        match = re.search(r"(?i)VID: (?P<vid>.*?),? ", show_inventory_chassis, re.MULTILINE)
        if match:
            self._udi['vid'] = match.group('vid')

        match = re.search(r"(?i)SN: (?P<sn>.*)", show_inventory_chassis, re.MULTILINE)
        if match:
            self._udi['sn'] = match.group('sn')

    def discovery(self, logfile=None):
        """This method detects the device details. This method discovery the several device attributes.

        Args:
            logfile (file): Optional file descriptor for session logging. The file must be open for write.
                The session is logged only if ``log_session=True`` was passed to the constructor.
                It the parameter is not passed then the default *session.log* file is created in `log_dir`.

        """

        self._set_default_log_fd(logfile)

        self._init_driver()

        no_hosts = len(self._nodes)
        for _ in xrange(no_hosts):
            try:
                self._driver.connect(logfile=self._session_fd)
                break
            except ConnectionError:
                self._shift_driver()
        else:
            raise ConnectionError("Unable to connect to the device")

        self._update_device_info()
        self._update_udi()

        self._prompt = self._driver.prompt
        self._is_console = self._detect_console()
        self._driver.disconnect()

        driver_name = self._get_driver_name()
        if driver_name == 'generic':
            raise RuntimeError("Platform {} not supported".format(self.family))

        self._init_driver(driver_name)
        self._driver.determine_hostname(self._prompt)

        self._hostname = self._driver.hostname

        self.logger.info("Hostname: '{}'".format(self.hostname))
        self.logger.info("Family: {}".format(self.family))
        self.logger.info("Platform: {}".format(self.platform))
        self.logger.info("OS: {}".format(os_names[self.os_type]))
        self.logger.info("Version: {}".format(self.os_version))
        self.logger.info("Name: {}".format(self.udi['name']))
        self.logger.info("Description: {}".format(self.udi['description']))
        self.logger.info("PID: {}".format(self.udi['pid']))
        self.logger.info("VID: {}".format(self.udi['vid']))
        self.logger.info("SN: {}".format(self.udi['sn']))
        self.logger.info("Prompt: '{}'".format(self._prompt))
        self.logger.info("Is connected to console: {}".format(self.is_console))

    def store_property(self, key, value):
        """This method stores a *value* identified by the *key* in the :class:`Connection` object.

        Args:
            key (str): Key name.
            value (object): Object to be stored

        """

        self.logger.debug("Store '{}' <- '{}'".format(key, str(value)))
        self._info[key] = value

    def get_property(self, key):
        """This method retrieves value of the *key* stored in :class:`Connection` object.
        If no value was stored for give key then None is returned

        Args:
            key (str): Key name.
        Returns:
            object|None: The stored object identified by key or None
        """

        return self._info.get(key, None)

    def connect(self, logfile=None):
        """This method connects to the device. The discovery method must be called first. If not then
        :class:`ConnectionError` is raised

        Args:
            logfile (file): Optional file descriptor for session logging. The file must be open for write.
                The session is logged only if ``log_session=True`` was passed to the constructor.
                It the parameter is not passed then the default *session.log* file is created in `log_dir`.

        Raises:
            ConnectionError: If the discovery method was not called first or there was a problem with getting
             the connection.
            ConnectionAuthenticationError: If the authentication failed.
            ConnectionTimeoutError: If the connection timeout happened.

        """
        self._set_default_log_fd(logfile)
        self._init_driver()
        no_hosts = len(self._nodes)
        result = False
        for i in xrange(no_hosts):
            try:
                result = self._driver.connect(logfile=self._session_fd)
                break
            except ConnectionError as e:
                # if this is last try raise the exception
                if (i + 1) == no_hosts:
                    raise e
                else:
                    self._shift_driver()
            except AttributeError:
                raise ConnectionError("Platform unknown. Try detect platform first")

        else:
            # This will never be executed
            raise ConnectionError("Unable to connect to the device")

        return result

    def reconnect(self, max_timeout=360, logfile=None):
        """This method reconnects to the device. It can be called when after device reloads or the session was
        disconnected either by device or jumphost. If multiple jumphosts are used then `reconnect` starts from
        the last valid connection.

        Args:
            max_timeout (int): This is the maximum amount of time during the session tries to reconnect. It may take
                longer depending on the TELNET or SSH default timeout.
            logfile (file): Optional file descriptor for session logging. The file must be open for write.
                The session is logged only if ``log_session=True`` was passed to the constructor.
                It the parameter is not passed then the default *session.log* file is created in `log_dir`.

        Raises:
            ConnectionError: If the discovery method was not called first or there was a problem with getting
             the connection.
            ConnectionAuthenticationError: If the authentication failed.
            ConnectionTimeoutError: If the connection timeout happened.

        """
        self._set_default_log_fd(logfile)

        begin = time.time()
        expired = 0.0
        attempt = 0
        self.logger.info("Trying to reconnect within {} seconds".format(max_timeout))
        while expired < max_timeout:
            self.logger.debug("Reconnecting. Attempt {}".format(attempt))
            try:
                self._driver.reconnect(logfile=self._session_fd)
                break
            except ConnectionError:
                expired = time.time() - begin
                self._shift_driver()
            except AttributeError:
                raise ConnectionError("Platform unknown. Try detect platform first")
            attempt += 1
        else:
            self.logger.error("Unable to reconnect")
            raise ConnectionTimeoutError("Unable to reconnect to device within {} s".format(max_timeout))

    def disconnect(self):
        """
        This method disconnect the session from the device and all the jumphosts in the path.
        """
        try:
            self._driver.disconnect()
        except AttributeError:
            pass
        finally:
            if self._session_fd:
                self._session_fd = None

    @property
    def platform(self):
        """Returns the string representing hardware platform model. For example: ASR-9010, ASR922, NCS-4006, etc."""
        __pid = self._udi['pid']
        match = re.search(r"([A-Z]{3}[-| ]?[0-9]{3,4})", __pid)
        if match:
            return match.group(1)
        else:
            return self._platform

    @platform.setter
    def platform(self, platform):
        self._platform = platform

    @property
    def family(self):
        """Returns the string representing hardware platform family. For example: ASR9K, ASR900, NCS6K, etc."""
        #if self._family == 'generic':
        #    self.discovery()
        return self._family

    @property
    def prompt(self):
        """Returns the target device prompt if detected or *None*."""
        try:
            return self._driver.prompt
        except AttributeError:
            return None

    @property
    def os_type(self):
        """Returns the string representing the Operating System type. For example: IOS, XR, eXR. If not detected returns
        *None*"""
        try:
            return self._os_type
        except AttributeError:
            return None

    @property
    def os_version(self):
        """Returns the string representing the Operating System version. For example 5.3.1.
        If not detected returns *None*"""
        try:
            return self._os_version
        except AttributeError:
            return None

    @property
    def hostname(self):
        """Returns the string representing the target device hostname."""
        try:
            return self._hostname
        except AttributeError:
            return "host"

    @property
    def is_connected(self):
        """Returns boolean value. *True* if target device is connected, *False* if not connected"""
        try:
            return self._driver.is_connected
        except AttributeError:
            return False

    @property
    def is_console(self):
        """Returns *True* if the connection to the target device is over console port"""
        return self._is_console

    @property
    def name(self):
        """Returns the chassis name"""
        return self._udi['name']

    @property
    def description(self):
        """Returns the chassis description."""
        return self._udi['description']

    @property
    def pid(self):
        """Returns the chassis PID."""
        return self._udi['pid']

    @property
    def vid(self):
        """Returns the chassis VID."""
        return self._udi['vid']

    @property
    def sn(self):
        """Returns the chassis SN."""
        return self._udi['sn']

    @property
    def udi(self):
        """Returns the dict representing the udi hardware record::

            {'description': 'ASR-9904 AC Chassis', 'name': 'Rack 0', 'pid': 'ASR-9904-AC', 'sn': 'FOX1830GT5W ', 'vid': 'V01'}

        """
        return self._udi

    @property
    def device_info(self):
        """Returns the dict represeing the device info record::

            {'family': 'ASR9K', 'os_type': 'eXR', 'os_version': '6.1.0.06I', 'platform': 'ASR-9904'}

        """
        _device_info = {
            'family': self.family,
            'platform': self.platform,
            'os_type': self.os_type,
            'os_version': self.os_version
        }
        return _device_info



# =============================================================================
# asr9k
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import logging
import re

import generic
from ..exceptions import ConnectionError, ConnectionAuthenticationError

from ..controllers.fsm import FSM, action
import pexpect


_logger = logging.getLogger(__name__)


class Connection(generic.Connection):
    """
    This is a platform specific implementation of based Driver class
    """
    platform = 'IOS'
    command_syntax_re = re.compile('\% Bad IP address or host name% Unknown command or computer name, '
                                   'or unable to find computer address|'
                                   '\% Ambiguous command:.*"|'
                                   '\% Type "show \?" for a list of subcommands|'
                                   "% Invalid input detected at '\^' marker\.")

    platform_prompt = generic.prompt_patterns['IOS']

    password_prompt = re.compile("Password: ")
    username_prompt = re.compile("Username: ")
    rommon_prompt = re.compile("(rommon \d+ >)|(rommon>)")

    def _get_enable_password(self):
        hop_info = self.hosts[-1]
        enable_password = hop_info.enable_password
        if enable_password is None:
            enable_password = hop_info.password
        return enable_password

    def prepare_prompt(self):
        mode = self.ctrl.detected_target_prompt[-1]

        # previously: '({})(\([^()]*\))?[#|>]'
        # It is assumed that session stays in privilege mode
        prompt_re = re.compile('({})(\([^()]*\))?#'.format(
            re.escape(self.ctrl.detected_target_prompt[:-1])))
        self.compiled_prompts[-1] = prompt_re
        self.prompt = self.ctrl.detected_target_prompt

        if mode == '>':
            self.enable()

    def determine_hostname(self, prompt):
        result = re.search(r"^(.*)[#|>]", prompt)
        if result:
            self.hostname = result.group(1)
            self._debug("Hostname detected: {}".format(self.hostname))

    def prepare_terminal_session(self):
        self.send('terminal len 0')
        self.send('terminal width 0')

    def reload(self, save_config=True):
        """
        CSM_DUT#reload

        System configuration has been modified. Save? [yes/no]: yes
        Building configuration...
        [OK]
        Proceed with reload? [confirm]
        """
        pass

    def enable(self, enable_password=None):
        ENABLE = "enable"
        self.ctrl.send("enable")
        prompt = self.compiled_prompts[-1]

        events = [ENABLE, self.password_prompt, prompt, pexpect.TIMEOUT, pexpect.EOF]
        transitions = [
            (ENABLE, [0], 1, self._send_lf, 10),
            (self.password_prompt, [1], 2, self._send_enable_password, 10),
            (self.password_prompt, [2], -1,
             ConnectionAuthenticationError("Incorrect enable password", self.hostname), 0),
            (prompt, [1, 2, 3], -1, None, 0),
            (pexpect.TIMEOUT, [0, 1, 2], -1,
             ConnectionAuthenticationError("Unable to get enable mode", self.hostname), 0),
            (pexpect.EOF, [0, 1, 2], -1, ConnectionError("Device disconnected", self.hostname), 0)
        ]

        fs = FSM("IOS-ENABLE", self.ctrl, events, transitions, timeout=10)
        return fs.run()

    @action
    def _send_lf(self, ctx):
        ctx.ctrl.send('\r')
        return True

    @action
    def _send_enable_password(self, ctx):
        ctx.ctrl.setecho(False)
        ctx.ctrl.sendline(self._get_enable_password())
        ctx.ctrl.setecho(True)
        return True



# =============================================================================
# ssh
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

from base import *
import pexpect
from ..fsm import FSM, action

from ...exceptions import \
    ConnectionAuthenticationError, \
    ConnectionError, \
    ConnectionTimeoutError


MODULUS_TOO_SMALL = "modulus too small"
PROTOCOL_DIFFER = "Protocol major versions differ"
NEWSSHKEY = "fingerprint is"
KNOWN_HOSTS = "added.*to the list of known hosts"
HOST_KEY_FAILED = "key verification failed"


class SSH(Protocol):
    def __init__(self, controller, device, spawn, account_manager, logfile):
        super(SSH, self).__init__(controller, device, account_manager, logfile)

        command = self._get_command()
        if spawn:
            self._spawn_session(command)

    def _get_command(self, version=2):
        if self.username:
            command = "ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -{} -p {} {}@{}".format(
                version, self.port, self.username, self.hostname
            )
        else:
            command = "ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no -{} -p {} {}".format(
                version, self.port, self.hostname
            )
        return command

    def connect(self):

        if self.ctrl.is_target:
            prompt = self.ctrl.platform.platform_prompt
            rommon_prompt = self.ctrl.platform.rommon_prompt
            more = self.ctrl.platform.more
            password_prompt = self.ctrl.platform.password_prompt
            username_prompt = self.ctrl.platform.username_prompt

        else:
            prompt = SHELL_PROMPT
            rommon_prompt = SHELL_PROMPT
            more = "!@#!@#"  # find another solution
            password_prompt = PASSWORD_PROMPT

        events = [password_prompt, prompt, UNABLE_TO_CONNECT, RESET_BY_PEER,
                  NEWSSHKEY, KNOWN_HOSTS, HOST_KEY_FAILED, MODULUS_TOO_SMALL, PROTOCOL_DIFFER,
                  pexpect.TIMEOUT]

        transitions = [
            (password_prompt, [0, 1, 4], -1, self.save_pattern, 0),
            (prompt, [0], -1, self.save_pattern, 0),
            #  cover all messages indicating that connection was not set up
            (UNABLE_TO_CONNECT, [0], -1, self.unable_to_connect, 0),
            #  not sure when it happens - saw if there was session timeout on router
            (RESET_BY_PEER, [0], -1, self.unable_to_connect, 0),
            (NEWSSHKEY, [0], 1, self.send_yes, 10),
            (KNOWN_HOSTS, [1], 0, None, 0),
            (HOST_KEY_FAILED, [0], -1, ConnectionError("Host key failed", self.hostname), 0),
            (MODULUS_TOO_SMALL, [0], 0, self.fallback_to_sshv1, 0),
            (PROTOCOL_DIFFER, [0], 4, self.fallback_to_sshv1, 0),
            (PROTOCOL_DIFFER, [4], -1, ConnectionError("Protocol version differs", self.hostname), 0),
            (pexpect.TIMEOUT, [0], 5, self.send_new_line, 10),
            (pexpect.TIMEOUT, [5], -1, ConnectionTimeoutError("Connection timeout", self.hostname), 0)

        ]
        sm = FSM("SSH-CONNECT", self.ctrl, events, transitions, timeout=30)
        return sm.run()

    def authenticate(self, prompt=None):
        if self.ctrl.is_target:
            prompt = self.ctrl.platform.prompt
            rommon_prompt = self.ctrl.platform.rommon_prompt
            password_prompt = self.ctrl.platform.password_prompt

        else:
            if prompt is None:
                prompt = SHELL_PROMPT
            rommon_prompt = SHELL_PROMPT
            password_prompt = PASSWORD_PROMPT

        events = [PRESS_RETURN, password_prompt, prompt, pexpect.TIMEOUT]

        transitions = [
            (PRESS_RETURN, [0, 1], 1, self.send_new_line, 10),
            (password_prompt, [0], 1, self.send_pass, 20),
            (password_prompt, [1], -1, ConnectionAuthenticationError("Authentication error", self.hostname), 0),
            (prompt, [0, 1], -1, None, 0),
            (pexpect.TIMEOUT, [1], -1,
             ConnectionError("Error getting device prompt") if self.ctrl.is_target else self.send_new_line, 0)
        ]

        if isinstance(prompt, str):
            self._dbg(10, "EXPECTED_PROMPT={}".format(prompt))
        elif prompt is not None:
            self._dbg(10, "EXPECTED_PROMPT={}".format(prompt.pattern))

        sm = FSM("SSH-AUTH", self.ctrl, events, transitions, init_pattern=self.last_pattern, timeout=30)
        sm.run()

        self.try_read_prompt(1)
        return True

    def disconnect(self):
        self.ctrl.sendline('\x03')

    @action
    def send_yes(self, ctx):
        ctx.ctrl.sendline("yes")
        return True

    @action
    def fallback_to_sshv1(self, ctx):
        command = self._get_command(version=1)
        self._spawn_session(command)
        return True

    def _dbg(self, level, msg):
        self.logger.log(
            level, "[{}]: [SSH]: {}".format(self.ctrl.hostname, msg)
        )


# =============================================================================
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import logging
import re

import generic
from ..exceptions import ConnectionError, ConnectionAuthenticationError

from ..controllers.fsm import FSM, action
import pexpect

_logger = logging.getLogger(__name__)


class Connection(generic.Connection):
    """
    This is a platform specific implementation of based Driver class
    """
    platform = 'NX-OS'
    command_syntax_re = re.compile("% Invalid command at '\^' marker\.|"
                                   "% Incomplete command at '\^' marker\.")

    platform_prompt = generic.prompt_patterns['NX-OS']

    password_prompt = re.compile("Password: ")
    username_prompt = re.compile("login: ")
    rommon_prompt = re.compile("loader >")
    standby_console = re.compile("\(standby\)")

    def prepare_prompt(self):
        mode = self.ctrl.detected_target_prompt[-1]

        # previously: '({})(\([^()]*\))?[#|>]'
        # It is assumed that session stays in privilege mode
        prompt_re = re.compile('({})(\([^()]*\))?#'.format(
            re.escape(self.ctrl.detected_target_prompt[:-1])))
        self.compiled_prompts[-1] = prompt_re
        self.prompt = self.ctrl.detected_target_prompt

    def determine_hostname(self, prompt):
        result = re.search(r"^(.*)#", prompt)
        if result:
            self.hostname = result.group(1)
            self._debug("Hostname detected: {}".format(self.hostname))

    def prepare_terminal_session(self):
        self.send('terminal len 0')
        self.send('terminal width 511')

    def reload(self, save_config=True):
        """
        !!!WARNING! there is unsaved configuration!!!
        This command will reboot the system. (y/n)?  [n]
        """
        if save_config:
            self.send("copy running-config startup-config")
        self.send("reload", wait_for_string="This command will reboot the system")
        self.ctrl.sendline("y")

    def enable(self, enable_password=None):
        pass



# =============================================================================
# generic.py
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


import re
import pexpect
from threading import Lock

from ..exceptions import \
    ConnectionError,\
    CommandSyntaxError, \
    CommandTimeoutError

from ..controllers.fsm import FSM, action

from ..controllers.protocols.base import PRESS_RETURN

_PROMPT_IOSXR = re.compile('\w+/\w+/\w+/\w+:.+#')
_PROMPT_SHELL = re.compile('\$\s*|>\s*')

_PROMPT_XML = 'XML> '
_INVALID_INPUT = "Invalid input detected"
_INCOMPLETE_COMMAND = "Incomplete command."
_CONNECTION_CLOSED = "Connection closed"

prompt_patterns = {
    'IOSXR': re.compile('(RP/\d+/RS?P[0-1]/CPU[0-3]:.*?)(\([^()]*\))?#'),
    'CALVADOS': re.compile("sysadmin-vm:[0-3]_RS?P[0-1]#"),
    'ROMMON': re.compile("rommon [A|B]?\d+ >"),
    'IOS': re.compile('[\w\-]+[#|>]'),
    'NX-OS': re.compile('[\w\-]+# '),
}

os_types = ['IOSXR', 'CALVADOS', 'ROMMON', 'IOS', 'NX-OS']


class Connection(object):
    platform = 'generic'
    shell_prompt = "\$\s?|>\s?|#\s?|\%\s?"
    connection_closed_re = re.compile("Connection closed")
    rommon_prompt = re.compile("rommon.*>")
    # platform_prompt = re.compile('[\r\n][\r\n]((\w+/\w+/\w+/\w+:.*?)(\([^()]*\))?#|.*?[#|>])')
    platform_prompt = re.compile('[\w\-]+[#>]')

    password_prompt = re.compile("[P|p]assword:\s?")
    username_prompt = re.compile("([U|u]sername:\s|login:\s?)")

    command_syntax_re = re.compile('\% Bad IP address or host name% Unknown command or computer name, '
                                   'or unable to find computer address|'
                                   '\% Ambiguous command:.*"|"'
                                   '\% Type "show \?" for a list of subcommands'
                                   '\%(w+)?for a list of subcommands|'
                                   '\% Ambiguous command:|'
                                   '\% Invalid input detected|'
                                   "\% Invalid command at .* marker")  # NX-OS
    press_return = re.compile("Press RETURN to get started\.")
    more = re.compile(" --More-- ")
    standby_console = re.compile("Standby console disabled|\(standby\)")

    def __init__(self, name, hosts, controller_class, logger, account_manager=None):
        self.hosts = hosts
        self.account_manager = account_manager
        self.pending_connection = False
        self.connected = False
        self.command_execution_pending = Lock()
        self.ctrl = None
        self.ctrl_class = controller_class
        self.name = name
        self.hostname = name
        self.logger = logger
        self.prompt = self.platform_prompt
        self._os_type = 'unknown'
        self.mode = None
        self.compiled_prompts = []
        for _ in hosts:
            self.compiled_prompts.append(None)

    def __repr__(self):
        name = ""
        for host in self.hosts:
            name += "->{}".format(host)
        return name[2:]

    def connect(self, logfile=None):
        """
        Connection initialization method.
        If logfile is None then the common logfile from
        Args:
            logfile (fd): File description for session log

        Returns:
            True if connection is established successfully
            False on failure.
        """

        if not self.connected:
            self.ctrl = self.ctrl_class(self, self.hostname, self.hosts, self.account_manager, logfile=logfile)
            self._info("Connecting to {} using {} driver".format(self.__repr__(), self.platform))
            self.connected = self.ctrl.connect()

        if self.connected:
            self._info("Connected to {}".format(self.__repr__()))
            self._compile_prompts()
            self.prepare_prompt()
            self.prepare_terminal_session()
        else:
            raise ConnectionError("Connection failed", self.hostname)

        return self.connected

    def disconnect(self):
        """
        Tear down the connection

        Args:
            None

        Returns:
            None

        """
        self._info("Disconnecting from {}".format(self.__repr__()))
        self.ctrl.disconnect()
        self.connected = False
        self.pending_connection = False
        self._info("Disconnected")

    def reconnect(self, logfile=None):
        self.connect(logfile=logfile)

    @property
    def is_connected(self):
        if self.connected:
            return self.ctrl.connected
        else:
            return False

    def send(self, cmd="", timeout=60, wait_for_string=None):
        """
        Send the command to the device and return the output

        Args:
            cmd (str): Command string for execution. Defaults to empty string.
            timeout (int): Timeout in seconds. Defaults to 60s
            wait_for_string (str): This is optional string that driver
                waits for after command execution. If none the detected
                prompt will be used.

        Returns:
            A string containing the command output.

        Raises:
            ConnectionError: General connection error during command execution
            CommandSyntaxError: Command syntax error or unknown command.
            CommandTimeoutError: Timeout during command execution
        """
        if self.connected:
            self._debug("Sending command: '{}'".format(cmd))

            try:
                self._execute_command(cmd, timeout, wait_for_string)
            except ConnectionError:
                self._warning("Connection lost. Disconnecting.")
                self.disconnect()
                raise

            self._info("Command executed successfully: '{}'".format(cmd))
            output = self.ctrl.before
            # if output.startswith(cmd):
            #    remove first line which contains the command itself
            #    second_line_index = output.find('\n') + 1
            #    output = output[second_line_index:]
            output = output.replace('\r', '')
            return output

        else:
            raise ConnectionError("Device not connected", host=self.hostname)

    def send_xml(self, command):
        """
        Handle error i.e.
        ERROR: 0x24319600 'XML-TTY' detected the 'informational' condition
        'The XML TTY Agent has not yet been started.
        Check that the configuration 'xml agent tty' has been committed.'
        """
        self._debug("Starting XML TTY Agent")
        result = self.send("xml", wait_for_string=_PROMPT_XML)
        if result != '':
            return result
        self._info("XML TTY Agent started")

        result = self.send(command, wait_for_string=_PROMPT_XML)
        self.ctrl.sendcontrol('c')
        self.send()
        return result

    def netconf(self, command):
        """
        Handle error i.e.
        ERROR: 0x24319600 'XML-TTY' detected the 'informational' condition
        'The XML TTY Agent has not yet been started.
        Check that the configuration 'xml agent tty' has been committed.'
        """
        self._debug("Starting XML TTY Agent")
        result = self.send("netconf", wait_for_string=']]>]]>')
        # if result != '':
        #    return result
        self._info("XML TTY Agent started")

        self.ctrl.send(command)
        self.ctrl.send("\r\n")
        self.ctrl.expect("]]>]]>")
        result = self.ctrl.before
        self.ctrl.sendcontrol('c')
        self.send()
        return result

    def enable(self, enable_password=None):
        """This method changes the device mode to privileged. If device does not support privileged mode the
        the informational message to the log will be posted.

        Args:
            enable_password (str): The privileged mode password. This is optional parameter. If password is not
                provided but required the password from url will be used. Refer to :class:`condoor.Connection`
        """
        self._info("Ignoring. Not supported on this platform")

    def reload(self, rommon_boot_command="boot"):
        """This method reloads the device and waits for device to boot up. It post the informational message to the
        log if not implemented by device driver."""

        self._info("Ignoring. Not implemented for this platform")

    def run_fsm(self, name, command, events, transitions, timeout, max_transitions=20):
        """This method instantiate and run the Finite State Machine for the current device connection. Here is the
        example of usage::

            test_dir = "rw_test"
            dir = "disk0:" + test_dir
            REMOVE_DIR = re.compile(re.escape("Remove directory filename [{}]?".format(test_dir)))
            DELETE_CONFIRM = re.compile(re.escape("Delete {}/{}[confirm]".format(filesystem, test_dir)))
            REMOVE_ERROR = re.compile(re.escape("%Error Removing dir {} (Directory doesnot exist)".format(test_dir)))

            command = "rmdir {}".format(dir)
            events = [device.prompt, REMOVE_DIR, DELETE_CONFIRM, REMOVE_ERROR, pexpect.TIMEOUT]
            transitions = [
                (REMOVE_DIR, [0], 1, send_newline, 5),
                (DELETE_CONFIRM, [1], 2, send_newline, 5),
                # if dir does not exist initially it's ok
                (REMOVE_ERROR, [0], 2, None, 0),
                (device.prompt, [2], -1, None, 0),
                (pexpect.TIMEOUT, [0, 1, 2], -1, error, 0)

            ]
            manager.log("Removing test directory from {} if exists".format(dir))
            if not device.run_fsm("DELETE_DIR", command, events, transitions, timeout=5):
                return False

        This FSM tries to remove directory from disk0:

        Args:
            name (str): Name of the state machine used for logging purposes. Can't be *None*
            command (str): The command sent to the device before FSM starts
            events (list): List of expected strings or pexpect.TIMEOUT exception expected from the device.
            transitions (list): List of tuples in defining the state machine transitions.
            timeout (int): Default timeout between states in seconds.
            max_transitions (int): Default maximum number of transitions allowed for FSM.

        The transition tuple format is as follows::

            (event, [list_of_states], next_state, action, timeout)

        - event (str): string from the `events` list which is expected to be received from device.
        - list_of_states (list): List of FSM states that triggers the action in case of event occurrence.
        - next_state (int): Next state for FSM transition.
        - action (func): function to be executed if the current FSM state belongs to `list_of_states` and the `event`
          occurred. The action can be also *None* then FSM transits to the next state without any action. Action
          can be also the exception, which is raised and FSM stops.

        The example action::

            def send_newline(ctx):
                ctx.ctrl.sendline()
                return True

            def error(ctx):
                ctx.message = "Filesystem error"
                return False

            def readonly(ctx):
                ctx.message = "Filesystem is readonly"
                return False

        The ctx object description refer to :class:`condoor.controllers.fsm.FSM`.

        If the action returns True then the FSM continues processing. If the action returns False then FSM stops
        and the error message passed back to the ctx object is posted to the log.


        The FSM state is the integer number. The FSM starts with initial ``state=0`` and finishes if the ``next_state``
        is set to -1.

        If action returns False then FSM returns False. FSM returns True if reaches the -1 state.

        """

        self._send_command(command)
        fsm = FSM(name, self.ctrl, events, transitions, timeout=timeout, max_transitions=max_transitions)
        return fsm.run()

    @property
    def os_type(self):
        return self._get_os_type()

    def _get_os_type(self):
        for os_type in os_types:
            prompt_pattern = prompt_patterns[os_type]
            if re.match(prompt_pattern, self.ctrl.detected_target_prompt):
                return os_type
        else:
            return "unknown"

    def prepare_terminal_session(self):
        self.send('terminal len 0')
        #self.send('terminal width 0')

    def _compile_prompts(self):
        self.compiled_prompts = [re.compile(re.escape(prompt)) for prompt in self.ctrl.detected_prompts]

    def _send_command(self, cmd):
        self.ctrl.setecho(False)
        self.ctrl.send(cmd)
        self.ctrl.expect_exact([cmd, pexpect.TIMEOUT], timeout=15)
        self.ctrl.sendline()
        self.ctrl.setecho(True)

    def _execute_command(self, cmd, timeout, wait_for_string):
        with self.command_execution_pending:
            try:
                self._send_command(cmd)

                if wait_for_string:
                    self._wait_for_string(wait_for_string, timeout)
                else:
                    self.wait_for_prompt(timeout)

            except CommandSyntaxError as e:
                self._error("{}: '{}'".format(e.message, cmd))
                e.command = cmd
                raise

            except (CommandTimeoutError, pexpect.TIMEOUT) as e:
                self._error("Command timeout: '{}'".format(cmd))
                raise CommandTimeoutError(message="Command timeout", host=self.hostname, command=cmd)

            except ConnectionError as e:
                self._error("{}: '{}'".format(e.message, cmd))
                raise

            except pexpect.EOF:
                self._error("Unexpected session disconnect")
                raise ConnectionError("Unexpected session disconnect", host=self.hostname)

            except Exception as err:
                error_msg = str(err)
                self._error("Exception: {}:{}".format(err.__class__, error_msg))
                raise ConnectionError(message=error_msg, host=self.hostname)

    def _determine_config_mode(self, prompt):
        if 'config' in prompt:
            self.mode = 'config'
        elif 'admin' in prompt:
            self.mode = 'admin'
        else:
            self.mode = 'global'

        self._debug("Mode: {}".format(self.mode))

    def determine_hostname(self, prompt):
        self._debug("Hostname detecting not implemented for generic driver")

    # Actions for FSM

    @action
    def _expected_string_received(self, ctx):
        ctx.finished = True
        return True

    @action
    def _expected_prompt(self, ctx):
        prompt = self.ctrl.after
        self.ctrl.detected_target_prompt = prompt
        self._determine_config_mode(prompt)
        self.determine_hostname(prompt)
        ctx.finished = True
        return True

    @action
    def _unexpected_prompt(self, ctx):
        prompt = self.ctrl.after
        ctx.msg = "Received the jump host prompt: '{}'".format(prompt)
        self.ctrl.last_hop = self.ctrl.detected_prompts.index(prompt)
        self.ctrl.connected = False
        return False

    @action
    def _connection_closed(self, ctx):
        prompt = self.ctrl.after
        ctx.msg = "Device disconnected."
        self.ctrl.connected = False
        # do not stop FSM to detect the jumphost prompt
        return True

    @action
    def print_before(self, ctx):
        print(":".join("{:02x}".format(ord(c)) for c in ctx.ctrl.before))
        print(ctx.ctrl.before)
        print(self.compiled_prompts[-1].pattern.encode('string_escape'))

    @action
    def _stays_connected(self, ctx):
        self.ctrl.connected = True
        self.ctrl.last_hop = len(self.ctrl.hosts) - 1  # Authentication needed
        self.ctrl.last_pattern = PRESS_RETURN
        return True

    @action
    def _send_lf(self, ctx):
        ctx.ctrl.send('\r')
        return True

    @action
    def _send_line(self, ctx):
        ctx.ctrl.send('\r\n')
        return True

    @action
    def _send_yes(self, ctx):
        ctx.ctrl.sendline('yes')
        return True

    @action
    def _send_space(self, ctx):
        ctx.ctrl.send(' ')
        return True

    def wait_for_prompt(self, timeout=60):
        events = [self.command_syntax_re, self.connection_closed_re,
                  pexpect.TIMEOUT, pexpect.EOF, self.compiled_prompts[-1], self.press_return, self.more]

        # add detected prompts chain
        events += self.compiled_prompts[:-1]  # without target prompt

        self._debug("Waiting for prompt")

        transitions = [
            (self.command_syntax_re, [0], -1, CommandSyntaxError("Command unknown", self.hostname), 0),
            (self.connection_closed_re, [0], 1, self._connection_closed, 10),
            (pexpect.TIMEOUT, [0], -1, CommandTimeoutError("Timeout waiting for prompt", self.hostname), 0),
            (pexpect.EOF, [0, 1], -1, ConnectionError("Unexpected device disconnect", self.hostname), 0),
            (self.more, [0], 0, self._send_space, 10),
            (self.compiled_prompts[-1], [0], -1, self._expected_prompt, 0),
            (self.press_return, [0], -1, self._stays_connected, 0)
        ]

        for prompt in self.compiled_prompts[:-1]:
            transitions.append((prompt, [0, 1], 0, self._unexpected_prompt, 0))

        sm = FSM("WAIT-4-PROMPT", self.ctrl, events, transitions, timeout=timeout)
        return sm.run()

    def _wait_for_string(self, expected_string, timeout=60):
        events = [self.command_syntax_re, self.connection_closed_re,
                  pexpect.TIMEOUT, pexpect.EOF, expected_string, self.press_return, self.more]

        # add detected prompts chain
        events += self.compiled_prompts[:-1]  # without target prompt

        self._debug("Waiting for string: '{}'".format(repr(expected_string)))

        transitions = [
            (self.command_syntax_re, [0], -1, CommandSyntaxError("Command unknown", self.hostname), 0),
            (self.connection_closed_re, [0], 1, self._connection_closed, 10),
            (pexpect.TIMEOUT, [0], -1, CommandTimeoutError("Timeout waiting for string", self.hostname), 0),
            (pexpect.EOF, [0, 1], -1, ConnectionError("Unexpected device disconnect", self.hostname), 0),
            (self.more, [0], 0, self._send_space, 10),
            (expected_string, [0], -1, self._expected_string_received, 0),
            (self.press_return, [0], -1, self._stays_connected, 0)
        ]

        for prompt in self.compiled_prompts[:-1]:
            transitions.append((prompt, [0, 1], 0, self._unexpected_prompt, 0))

        sm = FSM("WAIT-4-STR", self.ctrl, events, transitions, timeout=timeout)
        return sm.run()

    def prepare_prompt(self):
        self.prompt = self.ctrl.detected_target_prompt

    def _debug(self, msg):
        self.logger.debug("[{}]: {}".format(self.hostname, msg))

    def _error(self, msg):
        self.logger.error("[{}]: {}".format(self.hostname, msg))

    def _info(self, msg):
        self.logger.info("[{}]: {}".format(self.hostname, msg))

    def _warning(self, msg):
        self.logger.warning("[{}]: {}".format(self.hostname, msg))



# =============================================================================
# protocols
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================


from collections import defaultdict

from base import Protocol
from ssh import SSH
from telnet import Telnet

protocol2object = defaultdict(
    Protocol, {
        'ssh': SSH,
        'telnet': Telnet
    }
)


def make_protocol(controller, node_info, spawn, account_manager, logfile):
    return protocol2object[node_info.protocol](
        controller,
        node_info,
        spawn,
        account_manager,
        logfile)



# =============================================================================
# asr9k
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import re

import generic
import pexpect

from ..exceptions import \
    ConnectionError,\
    ConnectionAuthenticationError, \
    CommandSyntaxError, \
    CommandTimeoutError

from ..controllers.fsm import FSM, action
from ..controllers.protocols.base import RECONFIGURE_USERNAME_PROMPT


class Connection(generic.Connection):
    """
    This is a platform specific implementation of based Driver class
    """
    platform = 'ASR9K'
    command_syntax_re = re.compile("\%(w+)?for a list of subcommands|"
                                   "\% Ambiguous command:|"
                                   "\% Invalid input detected")

    platform_prompt = generic.prompt_patterns['IOSXR']

    password_prompt = re.compile("[Pp]assword: ")
    username_prompt = re.compile("[Uu]sername: ")
    rommon_prompt = re.compile("rommon \d+ >")
    calvados_prompt = re.compile("sysadmin-vm:[0-3]_RS?P[0-1]#")

    rommon_boot_command = "boot"

    def prepare_prompt(self):
        prompt_re = re.compile(
            '((({})(\([^()]*\))?|'
            'sysadmin-vm:[0-3]_RS?P[0-1]:?.*)#|'
            'rommon \d+ >|'
            'RP/[0-3]/RS?P[0-1]/CPU[0-1]:ios)'.format(re.escape(self.ctrl.detected_target_prompt[:-1]))
        )
        self.compiled_prompts[-1] = prompt_re
        self.prompt = self.ctrl.detected_target_prompt

    def prepare_terminal_session(self):
        self.send('terminal exec prompt no-timestamp')
        self.send('terminal len 0')
        self.send('terminal width 0')

    def determine_hostname(self, prompt):
        """
        RP/0/RP0/CPU0:Deploy#
        RP/0/RP0/CPU0:Deploy(config)#
        sysadmin-vm:0_RP0:NCS-Deploy2#
        sysadmin-vm:0_RP0:NCS-Deploy2(config)#
        sysadmin-vm:0_RP0#
        sysadmin-vm:0_RP0(config)#
        sysadmin-vm:0_RSP0#

        """
        try:
            if re.match(self.calvados_prompt, prompt):
                self.hostname = 'NOT-SET'
            else:
                self.hostname = prompt.split(":")[-1][:-1].split('(')[0]
            self._debug("Hostname detected: {}".format(self.hostname))
            if self.ctrl:
                self.ctrl.hostname = self.hostname
        except:
            raise
            self._warning("Unable to extract hostname from prompt: {}".format(prompt))


    def boot(self):
        pass

    def enable(self, enable_password=None):
        pass

    def reload(self, rommon_boot_command="boot", reload_timeout=300, os="XR"):
        """
        RP/0/RSP0/CPU0:ASR9K-PE4#reload
        Tue Nov 10 14:43:11.488 UTC
        Some active software packages are not yet committed. Proceed?[confirm]
        Standby card not present or not Ready for failover. Proceed?[confirm]
        Preparing system for backup. This may take a few minutes especially for large configurations.
        Status report: node0_RSP0_CPU0: START TO BACKUP
        Status report: node0_RSP0_CPU0: BACKUP HAS COMPLETED SUCCESSFULLY
        [Done]
        """

        self.rommon_boot_command = rommon_boot_command

        if os == "eXR":
            ADMIN = "admin"
            RELOAD = "hw-module location all reload"
            CONFIRM_RELOAD = re.compile(re.escape("Reload hardware module ? [no,yes]"))
            # STANDBY_CONSOLE = ios con0/RSP1/CPU0 is in standby
            STBY_CONSOLE = re.compile("ios con[0|1]/RS?P[0-1]/CPU[0-9] is in standby")
        else:
            RELOAD = "admin reload location all"
            PROCEED = re.compile(re.escape("Proceed with reload? [confirm]"))

        DONE = re.compile(re.escape("[Done]"))
        CONFIGURATION_COMPLETED = re.compile("SYSTEM CONFIGURATION COMPLETED")
        CONFIGURATION_IN_PROCESS = re.compile("SYSTEM CONFIGURATION IN PROCESS")
        # CONSOLE = re.compile("ios con0/RSP0/CPU0 is now available"))
        CONSOLE = re.compile("ios con[0|1]/RS?P[0-1]/CPU0 is now available")

        RELOAD_NA = re.compile("Reload to the ROM monitor disallowed from a telnet line")

        # FIXME: Not sure need to set echo to false
        self.ctrl.setecho(False)

        transitions_shared = [
            # here must be authentication
            (CONSOLE, [3, 4], 5, None, 600),
            (self.press_return, [5], 6, self._send_lf, 300),
            # if asks for username/password reconfiguration, go to success state and let plugin handle the rest.
            (RECONFIGURE_USERNAME_PROMPT, [6, 7], -1, None, 0),
            (CONFIGURATION_IN_PROCESS, [6], 7, None, 180),
            (CONFIGURATION_COMPLETED, [7], -1, self._return_and_authenticate, 0),

            (pexpect.TIMEOUT, [0, 1, 2], -1,
             ConnectionAuthenticationError("Unable to reload", self.hostname), 0),
            (pexpect.EOF, [0, 1, 2, 3, 4, 5], -1,
             ConnectionError("Device disconnected", self.hostname), 0),
            (pexpect.TIMEOUT, [6], 7, self._send_line, 180),
            (pexpect.TIMEOUT, [7], -1,
             ConnectionAuthenticationError("Unable to reconnect after reloading", self.hostname), 0),
        ]

        if os == "eXR":
            self.send(cmd=ADMIN)
            self.ctrl.sendline(RELOAD)

            events = [RELOAD_NA, RELOAD, CONFIRM_RELOAD, DONE, STBY_CONSOLE, CONFIGURATION_IN_PROCESS, self.press_return,
                  CONSOLE, CONFIGURATION_COMPLETED, RECONFIGURE_USERNAME_PROMPT,
                  pexpect.TIMEOUT, pexpect.EOF]

            transitions = [
                # Preparing system for backup. This may take a few minutes especially for large configurations.
                (RELOAD, [0], 1, None, 120),
                (RELOAD_NA, [1], -1, self._reload_na, 0),
                (CONFIRM_RELOAD, [1], 2, self._send_yes, 120),
                (DONE, [2], 3, None, reload_timeout),
                (STBY_CONSOLE, [3], -1, None, 10)
            ] + transitions_shared

        else:
            self.ctrl.sendline(RELOAD)
            events = [RELOAD_NA, RELOAD, DONE, PROCEED, CONFIGURATION_IN_PROCESS, self.rommon_prompt, self.press_return,
                      CONSOLE, CONFIGURATION_COMPLETED, RECONFIGURE_USERNAME_PROMPT,
                      pexpect.TIMEOUT, pexpect.EOF]
            transitions = [
                # Preparing system for backup. This may take a few minutes especially for large configurations.
                (RELOAD, [0], 1, self._send_lf, 300),
                (RELOAD_NA, [1], -1, self._reload_na, 0),
                (DONE, [1], 2, None, 120),
                (PROCEED, [2], 3, self._send_lf, reload_timeout),
                (self.rommon_prompt, [0, 3], 4, self._send_boot, 600),
            ] + transitions_shared

        fs = FSM("RELOAD", self.ctrl, events, transitions, timeout=10)
        return fs.run()

    def wait_for_prompt(self, timeout=60):
        # ASR with IOSXR specific error when cmd is longer than 256 characters
        _BUFFER_OVERFLOW = "input buffer overflow"
        events = [self.command_syntax_re, self.connection_closed_re,
                  pexpect.TIMEOUT, pexpect.EOF, self.compiled_prompts[-1], self.press_return, self.more,
                  _BUFFER_OVERFLOW]

        # add detected prompts chain
        events += self.compiled_prompts[:-1]  # without target prompt

        self._debug("Waiting for prompt")

        transitions = [
            (self.command_syntax_re, [0], -1, CommandSyntaxError("Command unknown", self.hostname), 0),
            (self.connection_closed_re, [0], 1, self._connection_closed, 10),
            (pexpect.TIMEOUT, [0], -1, CommandTimeoutError("Timeout waiting for prompt", self.hostname), 0),
            (pexpect.EOF, [0, 1], -1, ConnectionError("Unexpected device disconnect", self.hostname), 0),
            (self.more, [0], 0, self._send_space, 10),
            (self.compiled_prompts[-1], [0, 1], -1, self._expected_prompt, 0),
            (self.press_return, [0], -1, self._stays_connected, 0),
            (_BUFFER_OVERFLOW, [0], -1, CommandSyntaxError("Command too long", self.hostname), 0)
        ]

        for prompt in self.compiled_prompts[:-1]:
            transitions.append((prompt, [0, 1], 0, self._unexpected_prompt, 0))

        sm = FSM("WAIT-4-PROMPT", self.ctrl, events, transitions, timeout=timeout)
        return sm.run()

    @action
    def _send_boot(self, ctx):
        ctx.ctrl.sendline(self.rommon_boot_command)
        return True

    @action
    def _authenticate(self, ctx):
        ctx.ctrl.connect(start_hop=len(ctx.ctrl.hosts)-1, spawn=False, detect_prompt=False)
        return True

    @action
    def _return_and_authenticate(self, ctx):
        self._send_lf(ctx)
        ctx.ctrl.connect(start_hop=len(ctx.ctrl.hosts)-1, spawn=False, detect_prompt=False)
        return True

    @action
    def _reload_na(self, ctx):
        ctx.msg = "Reload to the ROM monitor disallowed from a telnet line. " \
                  "Set the configuration register boot bits to be non-zero."
        ctx.failed = True
        return False


# =============================================================================
# fsm
#
# Copyright (c)  2016, Cisco Systems
# All rights reserved.
#
# # Author: Klaudiusz Staniek
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
# =============================================================================

import logging
from functools import wraps
from pexpect import EOF
from time import time

from ..exceptions import \
    ConnectionError


def action(func):
    @wraps(func)
    def with_logging(*args, **kwargs):
        for arg in args:
            if isinstance(arg, FSM.Context):
                logging.getLogger(__name__).debug("[{}]: [{}] A={}".format(
                    arg.ctrl.hostname, arg.fsm_name, func.__name__))
                break
        else:
            logging.getLogger(__name__).debug("[FSM] A={}".format(func.__name__))
        return func(*args, **kwargs)
    return with_logging


class FSM(object):
    """This class represents Finite State Machine for the current device connection. Here is the
        example of usage::

            to be done


        The example action::

            def send_newline(ctx):
                ctx.ctrl.sendline()
                return True

            def error(ctx):
                ctx.message = "Filesystem error"
                return False

            def readonly(ctx):
                ctx.message = "Filesystem is readonly"
                return False

        The ctx object description refer to :class:`condoor.controllers.fsm.FSM`.

        If the action returns True then the FSM continues processing. If the action returns False then FSM stops
        and the error message passed back to the ctx object is posted to the log.


        The FSM state is the integer number. The FSM starts with initial ``state=0`` and finishes if the ``next_state``
        is set to -1.

        If action returns False then FSM returns False. FSM returns True if reaches the -1 state.

        """

    class Context(object):
        _slots__ = ['fsm_name', 'ctrl', 'event_index', 'event', 'state', 'finished', 'msg']
        fsm_name = "FSM"
        ctrl = None
        event_index = 0
        event = None
        state = 0
        finished = False
        msg = ""

        def __init__(self, fsm_name, ctrl):
            """This is a class constructor.

            Args:
                fsm_name (str): Name of the FSM. This is used for logging.
                ctrl (object): The controller object.
            """
            self.ctrl = ctrl
            self.fsm_name = fsm_name

        def __str__(self):
            """Returns the string representing the context"""
            return "FSM Context:E={},S={},FI={},M='{}'".format(
                self.event, self.state, self.finished, self.msg)

    def __init__(self, name, ctrl, events, transitions, init_pattern=None, timeout=300, max_transitions=20):
        """This is a FSM class constructor.

        Args:
            name (str): Name of the state machine used for logging purposes. Can't be *None*
            ctrl (object): Controller class representing the connection to the device
            events (list): List of expected strings or pexpect.TIMEOUT exception expected from the device.
            transitions (list): List of tuples in defining the state machine transitions.
            init_pattern (str): The pattern that was expected in the previous operation.
            timeout (int): Timeout between states in seconds. Defaults to 300 seconds.
            max_transitions (int): Max number of transitions allowed before quiting the FSM.

        The transition tuple format is as follows::

            (event, [list_of_states], next_state, action, timeout)

        - event (str): string from the `events` list which is expected to be received from device.
        - list_of_states (list): List of FSM states that triggers the action in case of event occurrence.
        - next_state (int): Next state for FSM transition.
        - action (func): function to be executed if the current FSM state belongs to `list_of_states` and the `event`
          occurred. The action can be also *None* then FSM transits to the next state without any action. Action
          can be also the exception, which is raised and FSM stops.
        """
        self.events = events
        self.ctrl = ctrl
        self.timeout = timeout
        self.name = name
        self.init_pattern = init_pattern
        self.max_transitions = max_transitions
        self.logger = logging.getLogger('condoor.fsm')

        self.transition_table = self._compile(transitions, events)

    def _compile(self, transitions, events):
        compiled = {}
        for transition in transitions:
            event, states, new_state, action, timeout = transition
            if not isinstance(states, list):
                states = list(states)
            try:
                event_index = events.index(event)
            except ValueError:
                self._dbg(10, "Transition for non-existing event: {}".format(
                    event if isinstance(event, str) else event.pattern))
            else:
                for state in states:
                    key = (event_index, state)
                    compiled[key] = (new_state, action, timeout)

        return compiled

    def run(self):
        """This method starts the FSM.

            Returns:
                boolean: True if FSM reaches the last state or false if the exception or error message was raised
        """
        ctx = FSM.Context(self.name, self.ctrl)
        transition_counter = 0
        timeout = self.timeout
        self._dbg(10, "FSM Started")
        while transition_counter < self.max_transitions + 1:
            transition_counter += 1
            try:
                start_time = time()
                if self.init_pattern is None:
                    ctx.event = self.ctrl.expect(self.events, timeout=timeout)
                else:
                    if isinstance(self.init_pattern, str):
                        self._dbg(10, "INIT_PATTERN={}".format(self.init_pattern.encode('string_escape')))
                    elif self.init_pattern is not None:
                        self._dbg(10, "INIT_PATTERN={}".format(self.init_pattern.pattern.encode('string_escape')))
                    ctx.event = self.events.index(self.init_pattern)
                    self.init_pattern = None
                finish_time = time() - start_time
                key = (ctx.event, ctx.state)
                ctx.pattern = self.events[ctx.event]

                if key in self.transition_table:
                    transition = self.transition_table[key]
                    next_state, action, next_timeout = transition
                    self._dbg(10, "E={},S={},T={},RT={:.2f}".format(
                        ctx.event, ctx.state, timeout, finish_time))
                    if callable(action):
                        if not action(ctx):
                            self._dbg(50, "Error: {}".format(ctx.msg))
                            return False
                    elif isinstance(action, Exception):
                        raise action
                    elif action is None:
                        self._dbg(10, "No action")
                    else:
                        self._dbg(40, "FSM Action is not callable: {}".format(action.__name__))
                        raise Exception("FSM Action is not callable")

                    if next_timeout != 0:  # no change if set to 0
                        timeout = next_timeout
                    ctx.state = next_state
                    self._dbg(10, "NS={},NT={}".format(next_state, timeout))

                else:
                    self._dbg(40, "Unknown transition: EVENT={},STATE={}".format(ctx.event, ctx.state))
                    continue

            except EOF:
                raise ConnectionError("Session closed unexpectedly", self.ctrl.hostname)

            if ctx.finished or next_state == -1:
                self._dbg(10, "FSM finished at E={},S={}".format(ctx.event, ctx.state))
                return True

        else:  # check while else if even exists
            self._dbg(40, "FSM looped. Exiting")
            return False

    def _dbg(self, level, msg):
        self.logger.log(
            level, "[{}]: [{}] {}".format(self.ctrl.hostname, self.name, msg)
        )
